use clap::Parser;
use std::{
    fs::{read_to_string, write},
    path::Path,
};
use toml::{Table, Value};

mod target;
use target::*;

mod config;
use config::*;

/// ngen is a build file generator for the ninja build system.
#[derive(Parser)]
struct Args {
    /// Path to the toml configuration file
    #[arg(short, long, default_value = "ngen.toml")]
    config_file: String,

    /// Path to generated ninja file
    #[arg(short, long, default_value = "build.ninja")]
    output_file: String,

    #[arg(long, hide = true)]
    write_compile_commands: Option<String>,
}

fn gen_ninja_code(target: &Target) -> String {
    let mut ret = String::from("\n");

    // necessary because you cannot access struct members in a format string :(
    let name = &target.name;
    let outfile = &target.outfile;
    let compiler = &target.compiler;
    let compiler_flags = &target.compiler_flags.join(" ");
    let linker = &target.linker;
    let linker_flags = &target.linker_flags.join(" ");
    let linker_libs = &target.linker_libs.join(" ");
    let sources = &target.sources;

    ret.push_str(&format!(
        "\
# BEGIN TARGET {name}
rule cc_{name}
  deps = gcc
  depfile = $dep
  command = {compiler} {compiler_flags} -MD -MF $dep -o $out -c $in
  description = Building $in -> $out
rule link_{name}
  command = {linker} {linker_flags} -o $out $in {linker_libs}
  description = Linking $out

build $builddir/{name}/obj: mkdir
build $builddir/{name}/dep: mkdir

"
    ));

    let mut object_list: Vec<String> = Vec::new();

    for s in sources {
        let new_file = s.replace("/", "-");
        let obj_name = format!("$builddir/{name}/obj/{new_file}.o");
        let dep_name = format!("$builddir/{name}/dep/{new_file}.o.d");
        ret.push_str(&format!(
            "build {obj_name}: cc_{name} {s}\n  dep = {dep_name}\n"
        ));
        object_list.push(obj_name);
    }

    ret.push_str(&format!("\nbuild $builddir/{name}/{outfile}: link_{name} "));
    ret.push_str(&object_list.join(" "));
    ret.push_str(&format!(
        " || $builddir/{name}/obj $builddir/{name}/dep\nbuild {name}: phony $builddir/{name}/{outfile}\n"
    ));
    if target.is_default {
        ret.push_str(&format!("\ndefault {name}\n"));
    }
    ret.push_str(&format!("# END TARGET {name}\n"));

    ret
}

fn gen_compile_commands(target: &Target, build_dir: &str) -> String {
    let mut ret = String::from("[\n");

    let name = &target.name;
    let compiler = &target.compiler;
    let compiler_flags = &target.compiler_flags.join(" ");

    let pwd = Path::new(".");
    let pwd = pwd
        .canonicalize()
        .unwrap_or_else(|e| panic!("cwd is non canonical (something is very wrong): {e}"));
    let pwd = pwd.to_str().unwrap();

    for (i, s) in target.sources.iter().enumerate() {
        let new_file = s.replace("/", "-");
        let obj_name = format!("{build_dir}/{name}/obj/{new_file}.o");

        ret.push_str(&format!(
            "  {{
    \"directory\": \"{pwd}\",
    \"command\": \"{compiler} {compiler_flags} -o {obj_name} -c {s}\",
    \"file\": \"{s}\"
  }}"
        ));
        if i != target.sources.len() - 1 {
            ret.push(',');
        }
        ret.push('\n');
    }

    ret.push_str("]\n");

    ret
}

fn main() {
    let args = Args::parse();
    let output_file = &args.output_file;
    let config_file_name = &args.config_file;

    let config_file = read_to_string(config_file_name)
        .unwrap_or_else(|e| panic!("could not read config file {config_file_name}: {e}"));
    let parsed_config = toml::from_str::<Table>(&config_file)
        .unwrap_or_else(|e| panic!("error parsing toml file {config_file_name}: {e}"));

    // get config
    let mut config = Config::get_default();
    for k in parsed_config.keys() {
        if let Value::Table(t) = parsed_config.get(k).unwrap() {
            if k == "config" {
                config.set(&t);
                break;
            }
        }
    }
    //----

    // parse the main target first
    let mut targets: Vec<Target> = Vec::new();
    let main_target = Target::new(&parsed_config, "main", None);
    targets.push(main_target.clone());
    let mut target_for_comp_cmds = main_target;
    let mut target_for_comp_cmds_changed = false;

    // parse other targets
    for k in parsed_config.keys() {
        if let Value::Table(t) = parsed_config.get(k).unwrap() {
            if k == "config" {
                continue; // ignore config
            }
            let target = Target::new(&t, k, Some(&targets[0]));
            targets.push(target);
            if let Some(_) = args.write_compile_commands {
                if *k == config.compile_commands_target {
                    target_for_comp_cmds = targets[targets.len() - 1].clone();
                    target_for_comp_cmds_changed = true;
                }
            }
        }
    }

    if let Some(file) = args.write_compile_commands {
        if config.compile_commands_target == "" {
            eprintln!(
                "ngen: warn: no compile_commands target found, using 'main' target by default."
            );
        } else if !target_for_comp_cmds_changed && config.compile_commands_target != "main" {
            let c = &config.compile_commands_target;
            eprintln!(
                "ngen: warn: compile_commands target {c} not found, using 'main' target instead."
            );
        }
        write(
            &file,
            &gen_compile_commands(&target_for_comp_cmds, &config.build_dir),
        )
        .unwrap_or_else(|e| panic!("fatal: could not write {file}: {e}"));
    }

    let build_dir = &config.build_dir;
    let mut ninjafile = format!(
        "\
# Generated by ngen. Do not modify by hand.

builddir = {build_dir}

rule mkdir
  command = mkdir -p $out
  description = Creating directory $out

rule regen_ninjafile
  command = ngen -c $in -o $out
  generator = 1
  description = Regenerating $out

",
    );

    if config.compile_commands {
        ninjafile.push_str(&format!(
            "\
rule regen_compile_commands
  command = ngen -c $in --write-compile-commands $out
  description = Regenerating $out

build $builddir: mkdir

build $builddir/compile_commands.json: regen_compile_commands {config_file_name} | $builddir
  pool = console

build {output_file}: regen_ninjafile {config_file_name} || $builddir/compile_commands.json
  pool = console
"
        ));
    } else {
        ninjafile.push_str(&format!(
            "build {output_file}: regen_ninjafile {config_file_name}\n  pool = console\n"
        ));
    }

    for t in targets {
        ninjafile.push_str(&gen_ninja_code(&t));
    }

    write(output_file, &ninjafile)
        .unwrap_or_else(|e| panic!("could not write ninja file {ninjafile}: {e}"));
}
